#ifndef BINARYSEARCHTREE_H
#define BINARYSEARCHTREE_H

#include "BinaryTree.h"

template <class dataType>
class BinarySearchTree
{
public:
    typedef BinaryTree<dataType> node;
    typedef BinarySearchTree<dataType> BST;

// ----------------------------------------------------------------
//  Name:           BinarySearchTree::BinarySearchTree
//  Description:    Default Constructor. an empty BST
//  Arguments:      - compare: A function which compares two
//                             dataTypes and returns -1 if the left
//                             is "less than" the right, 0 if they
//                             are equal, and 1 if the left is 
//                             "greater than" the right.
//  Return Value:   None.
// ----------------------------------------------------------------
    BinarySearchTree( int (*compare)(dataType& left, dataType& right) )
    {
        m_root = 0;
        m_count = 0;
        m_compare = compare;
    };


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::~BinarySearchTree
//  Description:    Destructor.
//  Arguments:      None.
//  Return Value:   None.
// ----------------------------------------------------------------
    ~BinarySearchTree()
    {
        if( m_root != 0 )
            delete m_root;
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::insert
//  Description:    Inserts data into the BST
//  Arguments:      - p_data: data to insert
//  Return Value:   None.
// ----------------------------------------------------------------
    void insert( dataType p_data )
    {
        node* current = m_root;

        if( m_root == 0 )
            m_root = new node( p_data );
        else
        {
            while( current != 0 )
            {
                if( m_compare( p_data, current->m_data ) < 0 )
                {
                    if( current->m_left == 0 )
                    {
                        current->m_left = new node( p_data );
                        current->m_left->m_parent = current;
                        current = 0;
                    }
                    else
                        current = current->m_left;
                }
                else
                {
                    if( current->m_right == 0 )
                    {
                        current->m_right = new node( p_data );
                        current->m_right->m_parent = current;
                        current = 0;
                    }
                    else
                        current = current->m_right;
                }
            }
        }
        m_count++;
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::has
//  Description:    finds the given data in the BST
//  Arguments:      - p_data: data to search for
//  Return Value:   true or false
// ----------------------------------------------------------------
    bool has( dataType p_data )
    {
        if( find( p_data ) == 0 )
            return false;
        return true;
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::get
//  Description:    finds the specified data within the tree and
//                  returns a pointer to it.
//  Arguments:      - p_data: data to search for
//  Return Value:   pointer to data, or 0 if not found.
// ----------------------------------------------------------------
    dataType* get( dataType p_data )
    {
        node* temp = find( p_data );
        if( temp != 0 )
            return &(temp->m_data);
        return 0;
    }

    
// ----------------------------------------------------------------
//  Name:           BinarySearchTree::remove
//  Description:    finds and removes data within the tree.
//  Arguments:      - p_data: data to find and remove
//  Return Value:   True if remove was successful, false if not.
// ----------------------------------------------------------------
    bool remove( dataType p_data )
    {
        node* x = find( p_data );
        if( x == 0 )
            return false;

        if( x->m_left == 0 || x->m_right == 0 )
            removeNotFull( x );
        else
            removeFull( x );
        m_count--;
        return true;
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::count
//  Description:    returns the count of items within the tree
//  Arguments:      None.
//  Return Value:   the number of items.
// ----------------------------------------------------------------
    inline int count()
    {
        return m_count;
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::depth
//  Description:    returns the depth of the root node
//  Arguments:      None.
//  Return Value:   depth of the root node.
// ----------------------------------------------------------------
    inline int depth()
    {
        if( m_root == 0 )
            return -1;
        return m_root->depth();
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::inorder
//  Description:    processes the BST using an inorder traversal.
//  Arguments:      - process: A function which takes a reference
//                             to a dataType and performs an
//                             operation on it.
//  Return Value:   None.
// ----------------------------------------------------------------
    inline void inorder( void (*process)(dataType& item) )
    {
        if( m_root != 0 )
            m_root->inorder( process );
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::inorderIterative
//  Description:    processes the BST using an iterative inorder 
//                  traversal.
//  Arguments:      - process: A function which takes a reference
//                             to a dataType and performs an
//                             operation on it.
//  Return Value:   None.
// ----------------------------------------------------------------
    inline void inorderIterative( void (*process)(dataType& item) )
    {
        node* current;
        current = m_root;
        if( current != 0 )
        {
            while( current->m_left != 0 )
                current = current->m_left;
            while(1)
            {
                process( current->m_data );
                if( current->m_right != 0 )
                {
                    current = current->m_right;
                    while( current->m_left != 0 )
                        current = current->m_left;
                }
                else
                {
                    if( current->isLeft() )
                    {
                        current = current->m_parent;
                    }
                    else
                    {
                        while( current->isRight() )
                            current = current->m_parent;
                        if( current->isRoot() )
                            return;
                        current = current->m_parent;
                    }
                }
            }
        }
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::preorder
//  Description:    processes the BST using a preorder traversal.
//  Arguments:      - process: A function which takes a reference
//                             to a dataType and performs an
//                             operation on it.
//  Return Value:   None.
// ----------------------------------------------------------------
    inline void preorder( void (*process)(dataType& item) )
    {
        if( m_root != 0 )
            m_root->preorder( process );
    }


protected:
// ----------------------------------------------------------------
//  Name:           BinarySearchTree::find
//  Description:    finds a piece of data in the tree and returns
//                  a pointer to the node which contains a match,
//                  or 0 if no match is found.
//  Arguments:      - p_data: data to find
//  Return Value:   pointer to node which contains data, or 0.
// ----------------------------------------------------------------
    node* find( dataType p_data )
    {
        node* current = m_root;
        int temp;
    
        while( current != 0 )
        {
            temp = m_compare( p_data, current->m_data );
            if( temp == 0 )
                return current;
            if( temp < 0 )
                current = current->m_left;
            else
                current = current->m_right;
        }
        return 0;
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::removeNotFull
//  Description:    removes a node from the tree which has only one
//                  or no children.
//  Arguments:      - x: pointer to the node to remove
//  Return Value:   None.
// ----------------------------------------------------------------
    void removeNotFull( node* x )
    {
        // find the valid child, or remain null if there
        // are no children.
        node* child = 0;
        if( x->m_left != 0 )
            child = x->m_left;
        if( x->m_right != 0 )
            child = x->m_right;

        if( x->isRoot() )
        {
            m_root = child;
        }
        else
        {
            if( x->isLeft() )
                x->m_parent->m_left = child;
            else
                x->m_parent->m_right = child;
        }

        // reset the parent node of the child, if it exists.
        if( child != 0 )
            child->m_parent = x->m_parent;

        x->m_left = 0;
        x->m_right = 0;
        delete x;
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::findLargestNode
//  Description:    finds the largest node in a subtree
//  Arguments:      - start: pointer to the node to start at
//  Return Value:   pointer to the largest node
// ----------------------------------------------------------------
node* findLargestNode( node* start )
{
    node* l = start;

    // travel the path all the way down to the right.
    // the rightmost node is the largest in the subtree.
    while( l->m_right != 0 )
    {
        l = l->m_right;
    }
    return l;
}


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::removeFull
//  Description:    removes a node from the tree which has two
//                  children
//  Arguments:      - x: pointer to the node to remove
//  Return Value:   None.
// ----------------------------------------------------------------
    void removeFull( node* x )
    {
        node* l = findLargestNode( x->m_left );

        // There are two cases:
        if( x->m_left == l )
        {
            // The largest node is immediately to the left
            // of the node to remove, so bypass x, and transfer
            // x's right child.
             l->m_right = x->m_right;
             l->m_right->m_parent = l;
        }
        else
        {
            // the largest node is somewhere deeper in the tree.
            // shortened removeNotFull algorithm <G>:
            // (notice how many assumptions can be made)
            l->m_parent->m_right = l->m_left;
            if( l->m_left != 0 )
                l->m_left->m_parent = l->m_parent;

            // now replace x with l.
            l->m_left = x->m_left;
            l->m_left->m_parent = l;
            l->m_right = x->m_right;
            l->m_right->m_parent = l;
        }

        if( x->isRoot() )
            m_root = l;
        else
        {
            if( x->isLeft() )
                x->m_parent->m_left = l;
            else
                x->m_parent->m_right = l;
        }

        l->m_parent = x->m_parent;
        x->m_left = 0;
        x->m_right = 0;
        delete x;
    }


// ----------------------------------------------------------------
//  Name:           BinarySearchTree::m_root
//  Description:    root node of the tree.
// ----------------------------------------------------------------
    node* m_root;

// ----------------------------------------------------------------
//  Name:           BinarySearchTree::m_count
//  Description:    count of variables within tree
// ----------------------------------------------------------------
    int m_count;

// ----------------------------------------------------------------
//  Name:           BinarySearchTree::m_compare
//  Description:    comparison function
// ----------------------------------------------------------------
    int (*m_compare)(dataType& left, dataType& right);
}; 

#endif
