#ifndef BINARYTREE_H
#define BINARYTREE_H

#include <iostream.h>

template <class dataType>
class BinaryTree
{
public:

    typedef BinaryTree<dataType> node;

// ----------------------------------------------------------------
//  Name:           BinaryTree::BinaryTree
//  Description:    Default Constructor. Creates a tree node.
//  Arguments:      - data: data to initialise node with
//  Return Value:   None.
// ----------------------------------------------------------------
    BinaryTree( dataType p_data )
        : m_data( p_data )
    {
        m_parent = 0;
        m_left   = 0;
        m_right  = 0;
    };


// ----------------------------------------------------------------
//  Name:           BinaryTree::~BinaryTree
//  Description:    Destructor. Deletes all child nodes.
//  Arguments:      None.
//  Return Value:   None.
// ----------------------------------------------------------------
    ~BinaryTree()
    {
        // delete both children. Since the destructor is recursive, 
        // each child node we delete automatically deletes all
        // of it's children. Neat, huh?
        if( m_left != 0 )
            delete m_left;
        if( m_right != 0 )
            delete m_right;
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::setLeft
//  Description:    sets the left child pointer, rearanging parents
//                  as neccesary. Existing left child is promoted
//                  to the root of it's own tree, but beware, it is
//                  up to the client to remember where it is, or
//                  else there is a memory leak.
//  Arguments:      - p_left: pointer to a binary tree node.
//  Return Value:   None.
// ----------------------------------------------------------------
    inline void setLeft( node* p_left )
    {
        if( m_left != 0 )
            m_left->m_parent = 0;
        m_left = p_left;
        p_left->m_parent = this;
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::setRight
//  Description:    sets the right child pointer, rearanging parents
//                  as neccesary. Existing right child is promoted
//                  to the root of it's own tree, but beware, it is
//                  up to the client to remember where it is, or
//                  else there is a memory leak.
//  Arguments:      - p_right: pointer to a binary tree node.
//  Return Value:   None.
// ----------------------------------------------------------------
    inline void setRight( node* p_right )
    {
        if( m_right != 0 )
            m_right->m_parent = 0;
        m_right = p_right;
        p_right->m_parent = this;
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::preorder
//  Description:    processes the tree using a pre-order traversal
//  Arguments:      - process: A function which takes a reference
//                             to a dataType and performs an
//                             operation on it.
//  Return Value:   None.
// ----------------------------------------------------------------
    void preorder( void (*process)(dataType& item) )
    {
        process( m_data );
        cout << " [ ";                      // debugging stuff.
        if( m_left != 0 )
            m_left->preorder( process );
        if( m_right != 0 )
            m_right->preorder( process );
        cout << " ] ";                      // debugging stuff.
    }

// ----------------------------------------------------------------
//  Name:           BinaryTree::postorder
//  Description:    processes the tree using a post-order traversal
//  Arguments:      - process: A function which takes a reference
//                             to a dataType and performs an
//                             operation on it.
//  Return Value:   None.
// ----------------------------------------------------------------
    void postorder( void (*process)(dataType& item) )
    {
        if( m_left != 0 )
            m_left->postorder( process );
        if( m_right != 0 )
            m_right->postorder( process );
        process( m_data );
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::inorder
//  Description:    processes the tree using an in-order traversal
//  Arguments:      - process: A function which takes a reference
//                             to a dataType and performs an
//                             operation on it.
//  Return Value:   None.
// ----------------------------------------------------------------
    void inorder( void (*process)(dataType& item) )
    {
        if( m_left != 0 )
            m_left->inorder( process );
        process( m_data );
        if( m_right != 0 )
            m_right->inorder( process );
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::isLeft
//  Description:    determines if node is a left subtree. Note that
//                  a result of false does NOT mean that it is a
//                  right child; it may be a root instead.
//  Arguments:      None.
//  Return Value:   True or False.
// ----------------------------------------------------------------
    inline bool isLeft()
    {
        if( isRoot() )
            return false;
        if( m_parent->m_left == this )
            return true;
        return false;
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::isRight
//  Description:    determines if node is a right subtree. Note that
//                  a result of false does NOT mean that it is a
//                  left child; it may be a root instead.
//  Arguments:      None.
//  Return Value:   True or False.
// ----------------------------------------------------------------
    inline bool isRight()
    {
        if( isRoot() )
            return false;
        if( m_parent->m_right == this )
            return true;
        return false;
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::isRoot
//  Description:    determines if node is a root node.
//  Arguments:      None.
//  Return Value:   True or False.
// ----------------------------------------------------------------
    inline bool isRoot()
    {
        if( m_parent == 0 )
            return true;
        return false;
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::count
//  Description:    recursively counts all children.
//  Arguments:      None.
//  Return Value:   the number of children.
// ----------------------------------------------------------------
    int count()
    {
        int temp = 1;
        if( m_left != 0 )
            temp += m_left->count();
        if( m_right != 0 )
            temp += m_right->count();
        return temp;
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::depth
//  Description:    recursively determines the lowest relative
//                  depth of all its children.
//  Arguments:      None.
//  Return Value:   the lowest depth of the current node.
// ----------------------------------------------------------------
    int depth()
    {
        int left = -1;
        int right = -1;
        if( m_left != 0 )
            left = m_left->depth();
        if( m_right != 0 )
            right = m_right->depth();
        if( left > right )
            return left + 1;
        return right + 1;
    }


// ----------------------------------------------------------------
//  Name:           BinaryTree::m_data
//  Description:    data stored in current tree node
// ----------------------------------------------------------------
    dataType m_data;

// ----------------------------------------------------------------
//  Name:           BinaryTree::m_parent
//  Description:    pointer to the parent of the node.
// ----------------------------------------------------------------
    node* m_parent;

// ----------------------------------------------------------------
//  Name:           BinaryTree::m_left
//  Description:    pointer to the left child of the node.
// ----------------------------------------------------------------
    node* m_left;

// ----------------------------------------------------------------
//  Name:           BinaryTree::m_right
//  Description:    pointer to the right child of the node.
// ----------------------------------------------------------------
    node* m_right;
}; 

#endif
